from collections.abc import AsyncIterator

import pytest_asyncio
from acp_sdk.client import Client

from e2e.config import Config


@pytest_asyncio.fixture
async def client() -> AsyncIterator[Client]:
    async with Client(base_url=f"http://localhost:{Config.PORT}") as client:
        yield client



from acp_sdk.client.types import Input
from acp_sdk.models.models import Message, MessagePart


def input_to_messages(input: Input) -> list[Message]:
    if isinstance(input, list):
        if len(input) == 0:
            return []
        if all(isinstance(item, Message) for item in input):
            return input
        elif all(isinstance(item, MessagePart) for item in input):
            return [Message(parts=input)]
        elif all(isinstance(item, str) for item in input):
            return [Message(parts=[MessagePart(content=content) for content in input])]
        else:
            raise TypeError("List with mixed types is not supported")
    else:
        if isinstance(input, str):
            input = MessagePart(content=input)
        if isinstance(input, MessagePart):
            input = Message(parts=[input])
        if isinstance(input, Message):
            input = [input]
        return input



import asyncio
import base64
from datetime import timedelta

import pytest
from acp_sdk.client import Client
from acp_sdk.models import (
    ArtifactEvent,
    ErrorCode,
    Message,
    MessageAwaitResume,
    MessageCreatedEvent,
    MessagePart,
    RunCompletedEvent,
    RunCreatedEvent,
    RunInProgressEvent,
    RunStatus,
)
from acp_sdk.models.errors import ACPError
from acp_sdk.server import Server

input = [Message(parts=[MessagePart(content="Hello!")])]
await_resume = MessageAwaitResume(message=Message(parts=[]))


@pytest.mark.asyncio
async def test_run_sync(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="echo", input=input)
    assert run.status == RunStatus.COMPLETED
    assert run.output == input


@pytest.mark.asyncio
async def test_run_async(server: Server, client: Client) -> None:
    run = await client.run_async(agent="echo", input=input)
    assert run.status == RunStatus.CREATED


@pytest.mark.asyncio
async def test_run_stream(server: Server, client: Client) -> None:
    event_stream = [event async for event in client.run_stream(agent="echo", input=input)]
    assert isinstance(event_stream[0], RunCreatedEvent)
    assert isinstance(event_stream[-1], RunCompletedEvent)


@pytest.mark.asyncio
async def test_run_status(server: Server, client: Client) -> None:
    run = await client.run_async(agent="echo", input=input)
    while run.status in (RunStatus.CREATED, RunStatus.IN_PROGRESS):
        run = await client.run_status(run_id=run.run_id)
    assert run.status == RunStatus.COMPLETED


@pytest.mark.asyncio
async def test_failure(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="failer", input=input)
    assert run.status == RunStatus.FAILED
    assert run.error is not None
    assert run.error.code == ErrorCode.INVALID_INPUT


@pytest.mark.asyncio
async def test_run_cancel(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="awaiter", input=input)
    assert run.status == RunStatus.AWAITING
    run = await client.run_cancel(run_id=run.run_id)
    assert run.status == RunStatus.CANCELLING


@pytest.mark.asyncio
async def test_run_resume_sync(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="awaiter", input=input)
    assert run.status == RunStatus.AWAITING
    assert run.await_request is not None

    run = await client.run_resume_sync(run_id=run.run_id, await_resume=await_resume)
    assert run.status == RunStatus.COMPLETED


@pytest.mark.asyncio
async def test_run_resume_async(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="awaiter", input=input)
    assert run.status == RunStatus.AWAITING
    assert run.await_request is not None

    run = await client.run_resume_async(run_id=run.run_id, await_resume=await_resume)
    assert run.status == RunStatus.IN_PROGRESS


@pytest.mark.asyncio
async def test_run_resume_stream(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="awaiter", input=input)
    assert run.status == RunStatus.AWAITING
    assert run.await_request is not None

    event_stream = [event async for event in client.run_resume_stream(run_id=run.run_id, await_resume=await_resume)]
    assert isinstance(event_stream[0], RunInProgressEvent)
    assert isinstance(event_stream[-1], RunCompletedEvent)


@pytest.mark.asyncio
async def test_run_session(server: Server, client: Client) -> None:
    async with client.session() as session:
        run = await session.run_sync(agent="echo", input=input)
        assert run.output == input
        run = await session.run_sync(agent="echo", input=input)
        assert run.output == input + input + input


@pytest.mark.asyncio
async def test_mime_types(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="mime_types", input=input)
    assert run.status == RunStatus.COMPLETED
    assert len(run.output) == 1

    message_parts = run.output[0].parts
    content_types = [part.content_type for part in message_parts]

    assert "text/html" in content_types
    assert "application/json" in content_types
    assert "application/javascript" in content_types
    assert "text/css" in content_types

    for part in message_parts:
        if part.content_type == "text/html":
            assert part.content == "<h1>HTML Content</h1>"
        elif part.content_type == "application/json":
            assert part.content == '{"key": "value"}'


@pytest.mark.asyncio
async def test_base64_encoding(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="base64_encoding", input=input)
    assert run.status == RunStatus.COMPLETED
    assert len(run.output) == 1

    message_parts = run.output[0].parts
    assert len(message_parts) == 2

    base64_part = next((part for part in message_parts if part.content_encoding == "base64"), None)
    assert base64_part is not None
    assert base64_part.content_type == "image/png"
    assert base64_part.content is not None

    text_part = next((part for part in message_parts if part.content_type == "text/plain"), None)
    assert text_part is not None
    assert text_part.content == "This is plain text"
    assert text_part.content_encoding == "plain"


@pytest.mark.asyncio
async def test_artifacts(server: Server, client: Client) -> None:
    run = await client.run_sync(agent="artifact_producer", input=input)
    assert run.status == RunStatus.COMPLETED

    assert len(run.output) == 1
    assert run.output[0].parts[0].content == "Processing with artifacts"

    assert len(run.output[0].parts) == 4

    text_artifact = next((a for a in run.output[0].parts if a.name == "text-result.txt"), None)
    json_artifact = next((a for a in run.output[0].parts if a.name == "data.json"), None)
    image_artifact = next((a for a in run.output[0].parts if a.name == "image.png"), None)

    assert text_artifact is not None
    assert text_artifact.content_type == "text/plain"
    assert text_artifact.content == "This is a text artifact result"
    assert text_artifact.content_encoding == "plain"

    assert json_artifact is not None
    assert json_artifact.content_type == "application/json"
    assert json_artifact.content == '{"results": [1, 2, 3], "status": "complete"}'

    assert image_artifact is not None
    assert image_artifact.content_type == "image/png"
    assert image_artifact.content_encoding == "base64"
    base64.b64decode(image_artifact.content)


@pytest.mark.asyncio
async def test_artifact_streaming(server: Server, client: Client) -> None:
    events = [event async for event in client.run_stream(agent="artifact_producer", input=input)]

    assert isinstance(events[0], RunCreatedEvent)
    assert isinstance(events[-1], RunCompletedEvent)

    message_events = [e for e in events if isinstance(e, MessageCreatedEvent)]
    artifact_events = [e for e in events if isinstance(e, ArtifactEvent)]

    assert len(message_events) == 1
    assert message_events[0].message.parts[0].content == "Processing with artifacts"

    assert len(artifact_events) == 3

    artifact_types = [a.part.content_type for a in artifact_events]
    assert "text/plain" in artifact_types
    assert "application/json" in artifact_types
    assert "image/png" in artifact_types


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [timedelta(seconds=5)], indirect=True)
async def test_run_ttl(server: Server, client: Client) -> None:
    run = await client.run_async(agent="echo", input=input)
    run = await client.run_status(run_id=run.run_id)
    await asyncio.sleep(6)
    try:
        run = await client.run_status(run_id=run.run_id)
        raise AssertionError("Error expected")
    except ACPError as e:
        if e.error.code == ErrorCode.NOT_FOUND:
            assert True
        else:
            raise AssertionError(f"Unexpected error code {e.error.code}")


@pytest.mark.asyncio
@pytest.mark.parametrize("server", [timedelta(seconds=5)], indirect=True)
async def test_session_ttl(server: Server, client: Client) -> None:
    async with client.session() as session:
        run = await session.run_sync(agent="echo", input=input)
        await asyncio.sleep(3)
        run = await session.run_sync(agent="echo", input=input)
        assert len(run.output) == 3
        await asyncio.sleep(3)
        run = await session.run_sync(agent="echo", input=input)
        assert len(run.output) == 7  # First run shall be forgotten



import asyncio
import os
from collections.abc import Awaitable
from datetime import timedelta
from typing import Any, Callable

import uvicorn
import uvicorn.config

from acp_sdk.models import Metadata
from acp_sdk.server.agent import Agent
from acp_sdk.server.agent import agent as agent_decorator
from acp_sdk.server.app import create_app
from acp_sdk.server.logging import configure_logger as configure_logger_func
from acp_sdk.server.telemetry import configure_telemetry as configure_telemetry_func


class Server:
    def __init__(self) -> None:
        self._agents: list[Agent] = []
        self._server: uvicorn.Server | None = None

    def agent(
        self,
        name: str | None = None,
        description: str | None = None,
        *,
        metadata: Metadata | None = None,
    ) -> Callable:
        """Decorator to register an agent."""

        def decorator(fn: Callable) -> Callable:
            agent = agent_decorator(name=name, description=description, metadata=metadata)(fn)
            self.register(agent)
            return fn

        return decorator

    def register(self, *agents: Agent) -> None:
        self._agents.extend(agents)

    def run(
        self,
        configure_logger: bool = True,
        configure_telemetry: bool = False,
        run_limit: int = 1000,
        run_ttl: timedelta = timedelta(hours=1),
        host: str = "127.0.0.1",
        port: int = 8000,
        uds: str | None = None,
        fd: int | None = None,
        loop: uvicorn.config.LoopSetupType = "auto",
        http: type[asyncio.Protocol] | uvicorn.config.HTTPProtocolType = "auto",
        ws: type[asyncio.Protocol] | uvicorn.config.WSProtocolType = "auto",
        ws_max_size: int = 16 * 1024 * 1024,
        ws_max_queue: int = 32,
        ws_ping_interval: float | None = 20.0,
        ws_ping_timeout: float | None = 20.0,
        ws_per_message_deflate: bool = True,
        lifespan: uvicorn.config.LifespanType = "auto",
        env_file: str | os.PathLike[str] | None = None,
        log_config: dict[str, Any]
        | str
        | uvicorn.config.RawConfigParser
        | uvicorn.config.IO[Any]
        | None = uvicorn.config.LOGGING_CONFIG,
        log_level: str | int | None = None,
        access_log: bool = True,
        use_colors: bool | None = None,
        interface: uvicorn.config.InterfaceType = "auto",
        reload: bool = False,
        reload_dirs: list[str] | str | None = None,
        reload_delay: float = 0.25,
        reload_includes: list[str] | str | None = None,
        reload_excludes: list[str] | str | None = None,
        workers: int | None = None,
        proxy_headers: bool = True,
        server_header: bool = True,
        date_header: bool = True,
        forwarded_allow_ips: list[str] | str | None = None,
        root_path: str = "",
        limit_concurrency: int | None = None,
        limit_max_requests: int | None = None,
        backlog: int = 2048,
        timeout_keep_alive: int = 5,
        timeout_notify: int = 30,
        timeout_graceful_shutdown: int | None = None,
        callback_notify: Callable[..., Awaitable[None]] | None = None,
        ssl_keyfile: str | os.PathLike[str] | None = None,
        ssl_certfile: str | os.PathLike[str] | None = None,
        ssl_keyfile_password: str | None = None,
        ssl_version: int = uvicorn.config.SSL_PROTOCOL_VERSION,
        ssl_cert_reqs: int = uvicorn.config.ssl.CERT_NONE,
        ssl_ca_certs: str | None = None,
        ssl_ciphers: str = "TLSv1",
        headers: list[tuple[str, str]] | None = None,
        factory: bool = False,
        h11_max_incomplete_event_size: int | None = None,
    ) -> None:
        if self._server:
            raise RuntimeError("The server is already running")

        import uvicorn

        if configure_logger:
            configure_logger_func()
        if configure_telemetry:
            configure_telemetry_func()

        config = uvicorn.Config(
            create_app(*self._agents, run_limit=run_limit, run_ttl=run_ttl),
            host,
            port,
            uds,
            fd,
            loop,
            http,
            ws,
            ws_max_size,
            ws_max_queue,
            ws_ping_interval,
            ws_ping_timeout,
            ws_per_message_deflate,
            lifespan,
            env_file,
            log_config,
            log_level,
            access_log,
            use_colors,
            interface,
            reload,
            reload_dirs,
            reload_delay,
            reload_includes,
            reload_excludes,
            workers,
            proxy_headers,
            server_header,
            date_header,
            forwarded_allow_ips,
            root_path,
            limit_concurrency,
            limit_max_requests,
            backlog,
            timeout_keep_alive,
            timeout_notify,
            timeout_graceful_shutdown,
            callback_notify,
            ssl_keyfile,
            ssl_certfile,
            ssl_keyfile_password,
            ssl_version,
            ssl_cert_reqs,
            ssl_ca_certs,
            ssl_ciphers,
            headers,
            factory,
            h11_max_incomplete_event_size,
        )
        self._server = uvicorn.Server(config)
        self._server.run()

    @property
    def should_exit(self) -> bool:
        return self._server.should_exit if self._server else False

    @should_exit.setter
    def should_exit(self, value: bool) -> None:
        self._server.should_exit = value



import asyncio
import logging
from collections.abc import AsyncGenerator
from concurrent.futures import ThreadPoolExecutor

from pydantic import BaseModel, ValidationError

from acp_sdk.instrumentation import get_tracer
from acp_sdk.models import (
    ACPError,
    AnyModel,
    AwaitRequest,
    AwaitResume,
    Error,
    ErrorCode,
    Event,
    GenericEvent,
    Message,
    MessageCompletedEvent,
    MessageCreatedEvent,
    MessagePart,
    MessagePartEvent,
    Run,
    RunAwaitingEvent,
    RunCancelledEvent,
    RunCompletedEvent,
    RunCreatedEvent,
    RunFailedEvent,
    RunInProgressEvent,
    RunStatus,
)
from acp_sdk.server.agent import Agent
from acp_sdk.server.logging import logger


class RunBundle:
    def __init__(
        self, *, agent: Agent, run: Run, input: list[Message], history: list[Message], executor: ThreadPoolExecutor
    ) -> None:
        self.agent = agent
        self.run = run
        self.input = input
        self.history = history

        self.stream_queue: asyncio.Queue[Event] = asyncio.Queue()

        self.await_queue: asyncio.Queue[AwaitResume] = asyncio.Queue(maxsize=1)
        self.await_or_terminate_event = asyncio.Event()

        self.task = asyncio.create_task(self._execute(input, executor=executor))

    async def stream(self) -> AsyncGenerator[Event]:
        while True:
            event = await self.stream_queue.get()
            if event is None:
                break
            yield event
            self.stream_queue.task_done()

    async def emit(self, event: Event) -> None:
        await self.stream_queue.put(event)

    async def await_(self) -> AwaitResume:
        await self.stream_queue.put(None)
        self.await_queue.empty()
        self.await_or_terminate_event.set()
        self.await_or_terminate_event.clear()
        resume = await self.await_queue.get()
        self.await_queue.task_done()
        return resume

    async def resume(self, resume: AwaitResume) -> None:
        self.stream_queue = asyncio.Queue()
        await self.await_queue.put(resume)
        self.run.status = RunStatus.IN_PROGRESS
        self.run.await_request = None

    async def cancel(self) -> None:
        self.task.cancel()
        self.run.status = RunStatus.CANCELLING
        self.run.await_request = None

    async def join(self) -> None:
        await self.await_or_terminate_event.wait()

    async def _execute(self, input: list[Message], *, executor: ThreadPoolExecutor) -> None:
        with get_tracer().start_as_current_span("run"):
            run_logger = logging.LoggerAdapter(logger, {"run_id": str(self.run.run_id)})

            in_message = False

            async def flush_message() -> None:
                nonlocal in_message
                if in_message:
                    await self.emit(MessageCompletedEvent(message=self.run.output[-1]))
                    in_message = False

            try:
                await self.emit(RunCreatedEvent(run=self.run))

                generator = self.agent.execute(
                    input=self.history + input, session_id=self.run.session_id, executor=executor
                )
                run_logger.info("Run started")

                self.run.status = RunStatus.IN_PROGRESS
                await self.emit(RunInProgressEvent(run=self.run))

                await_resume = None
                while True:
                    next = await generator.asend(await_resume)

                    if isinstance(next, (MessagePart, str)):
                        if isinstance(next, str):
                            next = MessagePart(content=next)
                        if not in_message:
                            self.run.output.append(Message(parts=[]))
                            in_message = True
                            await self.emit(MessageCreatedEvent(message=self.run.output[-1]))
                        self.run.output[-1].parts.append(next)
                        await self.emit(MessagePartEvent(part=next))
                    elif isinstance(next, Message):
                        await flush_message()
                        self.run.output.append(next)
                        await self.emit(MessageCreatedEvent(message=next))
                        for part in next.parts:
                            await self.emit(MessagePartEvent(part=part))
                        await self.emit(MessageCompletedEvent(message=next))
                    elif isinstance(next, AwaitRequest):
                        self.run.await_request = next
                        self.run.status = RunStatus.AWAITING
                        await self.emit(RunAwaitingEvent(run=self.run))
                        run_logger.info("Run awaited")
                        await_resume = await self.await_()
                        await self.emit(RunInProgressEvent(run=self.run))
                        run_logger.info("Run resumed")
                    elif isinstance(next, Error):
                        raise ACPError(error=next)
                    elif next is None:
                        await flush_message()
                    elif isinstance(next, BaseModel):
                        await self.emit(GenericEvent(generic=AnyModel(**next.model_dump())))
                    else:
                        try:
                            generic = AnyModel.model_validate(next)
                            await self.emit(GenericEvent(generic=generic))
                        except ValidationError:
                            raise TypeError("Invalid yield")
            except StopAsyncIteration:
                await flush_message()
                self.run.status = RunStatus.COMPLETED
                await self.emit(RunCompletedEvent(run=self.run))
                run_logger.info("Run completed")
            except asyncio.CancelledError:
                self.run.status = RunStatus.CANCELLED
                await self.emit(RunCancelledEvent(run=self.run))
                run_logger.info("Run cancelled")
            except Exception as e:
                if isinstance(e, ACPError):
                    self.run.error = e.error
                else:
                    self.run.error = Error(code=ErrorCode.SERVER_ERROR, message=str(e))
                self.run.status = RunStatus.FAILED
                await self.emit(RunFailedEvent(run=self.run))
                run_logger.exception("Run failed")
                raise
            finally:
                self.await_or_terminate_event.set()
                await self.stream_queue.put(None)



import logging

from opentelemetry import metrics, trace
from opentelemetry.exporter.otlp.proto.http._log_exporter import OTLPLogExporter
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import (
    SERVICE_NAME,
    SERVICE_NAMESPACE,
    SERVICE_VERSION,
    Resource,
)
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from acp_sdk.version import __version__

root_logger = logging.getLogger()


def configure_telemetry() -> None:
    """Utility that configures opentelemetry with OTLP exporter"""

    resource = Resource(
        attributes={
            SERVICE_NAME: "acp-server",
            SERVICE_NAMESPACE: "acp",
            SERVICE_VERSION: __version__,
        }
    )

    # Traces
    provider = TracerProvider(resource=resource)
    processor = BatchSpanProcessor(OTLPSpanExporter())
    provider.add_span_processor(processor)
    trace.set_tracer_provider(provider)

    # Metrics
    meter_provider = MeterProvider(
        resource=resource,
        metric_readers=[PeriodicExportingMetricReader(OTLPMetricExporter())],
    )
    metrics.set_meter_provider(meter_provider)

    # Logs
    logger_provider = LoggerProvider(resource=resource)
    processor = BatchLogRecordProcessor(OTLPLogExporter())
    logger_provider.add_log_record_processor(processor)
    root_logger.addHandler(LoggingHandler(logger_provider=logger_provider))



from concurrent.futures import ThreadPoolExecutor

import janus

from acp_sdk.models import SessionId
from acp_sdk.server.types import RunYield, RunYieldResume


class Context:
    def __init__(
        self,
        *,
        session_id: SessionId | None = None,
        executor: ThreadPoolExecutor,
        yield_queue: janus.Queue[RunYield],
        yield_resume_queue: janus.Queue[RunYieldResume],
    ) -> None:
        self.session_id = session_id
        self.executor = executor
        self._yield_queue = yield_queue
        self._yield_resume_queue = yield_resume_queue

    def yield_sync(self, value: RunYield) -> RunYieldResume:
        self._yield_queue.sync_q.put(value)
        return self._yield_resume_queue.sync_q.get()

    async def yield_async(self, value: RunYield) -> RunYieldResume:
        await self._yield_queue.async_q.put(value)
        return await self._yield_resume_queue.async_q.get()

    def shutdown(self) -> None:
        self._yield_queue.shutdown()
        self._yield_resume_queue.shutdown()



import pytest
from acp_sdk.client.types import Input
from acp_sdk.client.utils import input_to_messages
from acp_sdk.models import Message, MessagePart


@pytest.mark.parametrize(
    "input,messages",
    [
        ([], []),
        ("Hello", [Message(parts=[MessagePart(content="Hello")])]),
        (["Hello"], [Message(parts=[MessagePart(content="Hello")])]),
        (MessagePart(content="Hello"), [Message(parts=[MessagePart(content="Hello")])]),
        ([MessagePart(content="Hello")], [Message(parts=[MessagePart(content="Hello")])]),
        (Message(parts=[MessagePart(content="Hello")]), [Message(parts=[MessagePart(content="Hello")])]),
        ([Message(parts=[MessagePart(content="Hello")])], [Message(parts=[MessagePart(content="Hello")])]),
    ],
)
def test_input_to_messages(input: Input, messages: list[Message]) -> None:
    result = input_to_messages(input)
    assert result == messages


@pytest.mark.parametrize(
    "input",
    [["foo", Message(parts=[])], ["foo", MessagePart(content="foo")], [Message(parts=[]), MessagePart(content="foo")]],
)
def test_input_to_messages_mixed_input(input: Input) -> None:
    with pytest.raises(TypeError):
        input_to_messages(["foo", Message(parts=[])])



import uuid
from collections.abc import Iterator

from acp_sdk.models import Message, SessionId
from acp_sdk.models.models import RunStatus
from acp_sdk.server.bundle import RunBundle


class Session:
    def __init__(self) -> None:
        self.id: SessionId = uuid.uuid4()
        self.bundles: list[RunBundle] = []

    def append(self, bundle: RunBundle) -> None:
        self.bundles.append(bundle)

    def history(self) -> Iterator[Message]:
        for bundle in self.bundles:
            if bundle.run.status == RunStatus.COMPLETED:
                yield from bundle.input
                yield from bundle.run.output
